#if !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
#pragma once

// See NOTE: [Tensor vs. TensorBase]
namespace at {
class TensorBase;
}

namespace at::native {

TORCH_API bool cudnn_is_acceptable(const TensorBase& self);

} // namespace at::native

#else
#error "This file should not be included when either TORCH_STABLE_ONLY or TORCH_TARGET_VERSION is defined."
#endif  // !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
