#if !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
#pragma once
#include <torch/csrc/utils/python_stub.h>

// see [Note: Compiled Autograd]
namespace torch::dynamo::autograd {
PyObject* torch_c_dynamo_compiled_autograd_init();
} // namespace torch::dynamo::autograd

#else
#error "This file should not be included when either TORCH_STABLE_ONLY or TORCH_TARGET_VERSION is defined."
#endif  // !defined(TORCH_STABLE_ONLY) && !defined(TORCH_TARGET_VERSION)
