#pragma once

// ${generated_comment}

#include <ATen/core/Tensor.h>
#include <ATen/core/IListRef.h>
#include <ATen/TensorMeta.h>
#include <ATen/TensorIterator.h>

${NativeMetaFunctions_includes}

namespace at {

namespace meta {

${NativeMetaFunctions_declarations}

} // namespace meta
} // namespace at
